@@ -31,6 +31,20 @@
 #include <string.h>
 #include <assert.h>
 
+/*
+** For an build without mutexes, no-op the mutex calls.
+*/
+#if defined(SQLITE_THREADSAFE) && SQLITE_THREADSAFE==0
+#define sqlite3_mutex_alloc(X)    ((sqlite3_mutex*)8)
+#define sqlite3_mutex_free(X)
+#define sqlite3_mutex_enter(X)
+#define sqlite3_mutex_try(X)      SQLITE_OK
+#define sqlite3_mutex_leave(X)
+#define sqlite3_mutex_held(X)     ((void)(X),1)
+#define sqlite3_mutex_notheld(X)  ((void)(X),1)
+#endif /* SQLITE_THREADSAFE==0 */
+
+
 /************************ Object Definitions ******************************/
 
 /* Forward declaration of all object types */
@@ -309,7 +323,7 @@ static int quotaOpen(
           pFile=pFile->pNext){}
       if( pFile==0 ){
         int nName = strlen(zName);
-        pFile = sqlite3_malloc( sizeof(*pFile) + nName + 1 );
+        pFile = (quotaFile *)sqlite3_malloc( sizeof(*pFile) + nName + 1 );
         if( pFile==0 ){
           quotaLeave();
           pSubOpen->pMethods->xClose(pSubOpen);
@@ -669,7 +683,7 @@ int sqlite3_quota_set(
       quotaLeave();
       return SQLITE_OK;
     }
-    pGroup = sqlite3_malloc( sizeof(*pGroup) + nPattern + 1 );
+    pGroup = (quotaGroup *)sqlite3_malloc( sizeof(*pGroup) + nPattern + 1 );
     if( pGroup==0 ){
       quotaLeave();
       return SQLITE_NOMEM;
